#include <math.h>
#include <stdbool.h>
#include <cstdio>

#include <gint/gint.h>
#include <gint/display.h>
#include <gint/keyboard.h>
#include <gint/clock.h>
#include <libprof.h>
#include <string>
#include <vector>
#include "Color.hpp"
#include "Fixed.hpp"

#define MAP_W 8
#define MAP_H 8
#define TEXTURE_SIZE 40
#define FOV 60.0f

extern bopti_image_t wall_texture;
extern bopti_image_t ceil_texture;
extern bopti_image_t floor_texture;

ColorRaw inline GetColor(int x, int y, const bopti_image_t *texture) {
    uint8_t *data_u8 = (uint8_t *)((uint8_t *)texture->data + y * texture->stride);
    return texture->palette[((int8_t)data_u8[x])+128];
}

int map[MAP_H][MAP_W] = {
    {1,1,1,1,1,1,1,1},
    {1,0,0,0,0,0,0,1},
    {1,0,0,1,0,0,0,1},
    {1,0,0,1,0,0,0,1},
    {1,0,0,1,1,0,0,1},
    {1,0,0,0,0,0,0,1},
    {1,0,0,0,0,0,0,1},
    {1,1,1,1,1,1,1,1}
};

// Position joueur
fixed12_32 posX = 5.0f, posY = 4.0f;
fixed12_32 dir = 0.0f;

static std::vector<fixed12_32> rowDistLUT(DHEIGHT);
static ColorRaw shade_color[256];

void build_lut() {
    using fixed = fixed12_32;
    fixed halfH = fixed(DHEIGHT) / fixed(2);
    fixed k = fixed(DHEIGHT) / (-fixed(2));

    for (int y = 0; y < DHEIGHT; ++y) {
        fixed p = fixed(y) - halfH;
        if (p == fixed::zero()) p = fixed(0.001f);
        rowDistLUT[y] = k / p;
    }
}

void build_shade_color() {
    for (int i = 0; i < 256; ++i) {
        int shade = i;
        if (shade < 0) shade = 0;
        shade_color[i] = Color(shade, shade, shade).Raw();
    }
}

static void inline WritePixel(int x, int y, unsigned int color) {
    gint_vram[x + y * DWIDTH] = color;
}

bool inline is_wall(int mx, int my) {
    if (mx < 0 || mx >= MAP_W || my < 0 || my >= MAP_H) return true;
    return map[my][mx] != 0;
}

template <bool Shadow, bool Textured, bool FastFloorCeiling>
void cast_and_render() {
    using fixed = fixed12_32;
    const fixed PI_2 = fixed(MY_PI_2);
    const fixed FOV_RAD = fixed(FOV * MY_PI / 180.0);

    const fixed tanHalfFOV = fixed::tan(FOV_RAD / fixed(2));
    const fixed planeX_base = fixed::cos(dir + PI_2) * tanHalfFOV;
    const fixed planeY_base = fixed::sin(dir + PI_2) * tanHalfFOV;
    const fixed dirX = fixed::cos(dir);
    const fixed dirY = fixed::sin(dir);

    const fixed inv_SCREEN_W_base = fixed(2) * fixed::one() / fixed(DWIDTH);

    for (int x = 0; x < DWIDTH; x++) {
        fixed cameraX = fixed(x) * inv_SCREEN_W_base - fixed::one();

        fixed rayDirX = dirX + planeX_base * cameraX;
        fixed rayDirY = dirY + planeY_base * cameraX;

        int mapX = int(posX);
        int mapY = int(posY);

        fixed deltaDistX = (rayDirX == fixed::zero()) ? fixed::max_value() : fixed::one() / rayDirX.abs();
        fixed deltaDistY = (rayDirY == fixed::zero()) ? fixed::max_value() : fixed::one() / rayDirY.abs();

        fixed sideDistX, sideDistY;
        int stepX, stepY;
        int side = 0;

        if (rayDirX < fixed::zero()) {
            stepX = -1;
            sideDistX = (posX - fixed(mapX)) * deltaDistX;
        } else {
            stepX = 1;
            sideDistX = (fixed(mapX + 1) - posX) * deltaDistX;
        }

        if (rayDirY < fixed::zero()) {
            stepY = -1;
            sideDistY = (posY - fixed(mapY)) * deltaDistY;
        } else {
            stepY = 1;
            sideDistY = (fixed(mapY + 1) - posY) * deltaDistY;
        }

        while (true) {
            if (sideDistX < sideDistY) {
                sideDistX += deltaDistX;
                mapX += stepX;
                side = 0;
            } else {
                sideDistY += deltaDistY;
                mapY += stepY;
                side = 1;
            }
            if (is_wall(mapX, mapY)) break;
        }

        fixed perpWallDist = (side == 0) ? (sideDistX - deltaDistX) : (sideDistY - deltaDistY);
        if (perpWallDist <= fixed::zero()) perpWallDist = fixed(0.001f);

        int lineHeight = int(DHEIGHT / perpWallDist);

        int drawStart = -lineHeight / 2 + DHEIGHT / 2;
        int drawEnd   = lineHeight / 2 + DHEIGHT / 2;

        // Clamp uniquement le rendu, pas le lineHeight
        if (drawStart < 0) drawStart = 0;
        if (drawEnd > DHEIGHT) drawEnd = DHEIGHT;

        fixed wallX;
        if (side == 0) {
            wallX = posY + perpWallDist * rayDirY;
        } else {
            wallX = posX + perpWallDist * rayDirX;
        }
        wallX = wallX - wallX.floor();

        int texX = int(wallX * TEXTURE_SIZE);
        if (!((side == 0 && rayDirX > fixed::zero()) || (side == 1 && rayDirY < fixed::zero()))) {
            texX = TEXTURE_SIZE - texX - 1;
        }

        int shade = 0;
        if constexpr (Shadow || !Textured)
        {
            shade = 255 - int(perpWallDist * fixed(20));
            if (shade < 0) shade = 0;
        }
        if constexpr (Textured)
        {
            fixed texStep = fixed(TEXTURE_SIZE) / fixed(lineHeight);
            fixed texPos = (fixed(drawStart) - fixed(DHEIGHT) / fixed(2) + fixed(lineHeight) / fixed(2)) * texStep;
            ColorRaw texColor;
            for (int y = drawStart; y < drawEnd; y++)
            {
                if constexpr (Shadow)
                    texColor = Color::MultiplyColorRaw(GetColor(texX, texPos, &wall_texture), shade_color[shade]);
                else
                    texColor = GetColor(texX, texPos, &wall_texture);
                WritePixel(x, y, texColor);
                texPos += texStep;
            }
        }
        else
        {
            ColorRaw wallColor = Color(shade, shade, shade).Raw();
            for (int y = drawStart; y < drawEnd; y++)
                WritePixel(x, y, wallColor);
        }

        // --- SOL & PLAFOND TEXTURÉS ---
        // Prépare plan caméra
        if (FastFloorCeiling)
        {
            for (int y = 0; y < drawStart; y++)
            {
                WritePixel(x, y, COLOR_RED.Raw());                  // Plafond
                WritePixel(x, DHEIGHT - y - 1, COLOR_BLUE.Raw());    // Sol
            }
        }
        else
        {
            fixed planeX = planeX_base;
            fixed planeY = planeY_base;
            if constexpr (Textured)
            {
                fixed rayDirX0 = dirX - planeX;
                fixed rayDirY0 = dirY - planeY;
                fixed rayDirX1 = dirX + planeX;
                fixed rayDirY1 = dirY + planeY;

                // Pré-calcule fraction rayon horizontal
                fixed f = fixed(x) / fixed(DWIDTH);

                // Pré-calcule direction du rayon pour cette colonne
                fixed A = rayDirX0 + (rayDirX1 - rayDirX0) * f;
                fixed B = rayDirY0 + (rayDirY1 - rayDirY0) * f;

                for (int y = 0; y < drawStart; ++y) {
                    fixed rowDist = rowDistLUT[y];

                    fixed worldX = posX + rowDist * A;
                    fixed worldY = posY + rowDist * B;

                    fixed localX = worldX - worldX.floor();
                    fixed localY = worldY - worldY.floor();

                    int u = int(localX * TEXTURE_SIZE);
                    int v = int(localY * TEXTURE_SIZE);

                    ColorRaw uvColorCeil, uvColorFloor;
                    if constexpr (Shadow)
                    {
                        uvColorCeil = Color::MultiplyColorRaw(GetColor(u, v, &ceil_texture), shade_color[shade]);
                        uvColorFloor = Color::MultiplyColorRaw(GetColor(u, v, &floor_texture), shade_color[shade]);
                    }
                    else
                    {
                        uvColorCeil = GetColor(u, v, &ceil_texture);
                        uvColorFloor = GetColor(u, v, &floor_texture);
                    }

                    WritePixel(x, y, uvColorCeil);                  // Plafond
                    WritePixel(x, DHEIGHT - y - 1, uvColorFloor);    // Sol
                }
            }
            else
            {
                fixed A = dirX + planeX * cameraX;
                fixed B = dirY + planeY * cameraX;

                for (int y = 0; y < drawStart; ++y) {
                    fixed rowDist = rowDistLUT[y];
                    fixed ceilX = posX + rowDist * A;
                    fixed ceilY = posY + rowDist * B;

                    int tileX = int(ceilX.floor());
                    int tileY = int(ceilY.floor());
                    int checker = (tileX + tileY) & 1;

                    ColorRaw c1, c2;
                    if constexpr (Shadow)
                    {
                        shade = 255 - int(rowDist * fixed(TEXTURE_SIZE));
                        if (shade < 0) shade = 0;
                        if (checker) {
                            c1 = Color(shade, 0, 0).Raw();  // Rouge ombré
                            c2 = Color(0, 0, shade).Raw();  // Bleu ombré
                        } else {
                            c1 = Color(0, shade, 0).Raw();  // Vert ombré
                            c2 = Color(shade, shade, 0).Raw(); // Jaune ombré
                        }
                    }
                    else
                    {
                        if (checker) {
                            c1 = Color(255, 0, 0).Raw();  // Rouge
                            c2 = Color(0, 0, 255).Raw();  // Bleu
                        } else {
                            c1 = Color(0, 255, 0).Raw();  // Vert
                            c2 = Color(255, 255, 0).Raw(); // Jaune
                        }
                    }

                    WritePixel(x, y, c1);
                    WritePixel(x, DHEIGHT - y - 1, c2);
                }
            }
        }

    }
}

std::string float_to_string(float value) {
    long int intPart = (long int)value;
    long int fracPart = (long int)((value - intPart) * 100);
    if (fracPart < 0)
        fracPart = -fracPart;
    char buffer[50];
    snprintf(buffer, sizeof(buffer), "%ld.%02ld", intPart, fracPart);
    return std::string(buffer);
}

void cast_and_renderDynamic(bool shadow, bool textured, bool fastFloorCeiling) {
    if (shadow) {
        if (textured) {
            if (fastFloorCeiling)
                cast_and_render<true, true, true>();
            else
                cast_and_render<true, true, false>();
        } else {
            if (fastFloorCeiling)
                cast_and_render<true, false, true>();
            else
                cast_and_render<true, false, false>();
        }
    } else {
        if (textured) {
            if (fastFloorCeiling)
                cast_and_render<false, true, true>();
            else
                cast_and_render<false, true, false>();
        } else {
            if (fastFloorCeiling)
                cast_and_render<false, false, true>();
            else
                cast_and_render<false, false, false>();
        }
    }
}

int main() {
    prof_init();
    build_lut();
    build_shade_color();
    clock_set_speed(CLOCK_SPEED_F1);
    bool shadow = true;
    bool textured = true;
    bool fastFloorCeiling = false;
    bool overclock = false;
    while (!keydown(KEY_EXIT)) {
        prof_t perf_mafonction = prof_make();
        prof_enter_norec(perf_mafonction);
        //dclear(C_BLACK);

        cleareventflips();
        clearevents();
        fixed12_32 moveSpeed = 0.05f;
        fixed12_32 rotSpeed = 0.03f;

        fixed12_32 dirX = cosf(dir);
        fixed12_32 dirY = sinf(dir);
        if (keypressed(KEY_F1))
            shadow = !shadow;
        if (keypressed(KEY_F2))
            textured = !textured;
        if (keypressed(KEY_F3))
            fastFloorCeiling = !fastFloorCeiling;
        if (keypressed(KEY_F4))
        {
            overclock = !overclock;
            if (overclock) {
                clock_set_speed(CLOCK_SPEED_F5);
            } else {
                clock_set_speed(CLOCK_SPEED_F1);
            }
        }
        if (keydown(KEY_UP)) {
            fixed12_32 newX = posX + dirX * moveSpeed;
            fixed12_32 newY = posY + dirY * moveSpeed;
            if (!is_wall((int)newX, (int)newY)) {
                posX = newX;
                posY = newY;
            }
        }
        if (keydown(KEY_DOWN)) {
            fixed12_32 newX = posX - dirX * moveSpeed;
            fixed12_32 newY = posY - dirY * moveSpeed;
            if (!is_wall((int)newX, (int)newY)) {
                posX = newX;
                posY = newY;
            }
        }
        if (keydown(KEY_LEFT)) dir -= rotSpeed;
        if (keydown(KEY_RIGHT)) dir += rotSpeed;

        cast_and_renderDynamic(shadow, textured, fastFloorCeiling);

        prof_leave_norec(perf_mafonction);
        uint32_t perf1 = prof_time(perf_mafonction) + 1;
        char str[100];
        snprintf(str, sizeof(str), "%lu us (%lu fps)", perf1, 1000000 / perf1);
        dtext_opt(0, 0, C_WHITE, C_BLACK, DTEXT_LEFT, DTEXT_TOP, str, -1);

        snprintf(str, sizeof(str), "Pos: (%s, %s) Dir: %s rad",
                 float_to_string(posX).c_str(), float_to_string(posY).c_str(), float_to_string(dir).c_str());
        dtext_opt(0, 10, C_WHITE, C_BLACK, DTEXT_LEFT, DTEXT_TOP, str, -1);

        dupdate();
    }
    return 0;
}
